import json
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union, Callable, Optional

if TYPE_CHECKING:
    import pandas as pd
    from azure.kusto.data._models import KustoResultTable, KustoStreamingResultTable

# Alias for dataframe_from_result_table converter type
Converter = dict[str, Union[str, Callable[[str, "pd.DataFrame"], "pd.Series"]]]


def load_bundled_json(file_name: str) -> dict[Any, Any]:
    filename = Path(__file__).absolute().parent.joinpath(file_name)
    with filename.open("r", encoding="utf-8") as data:
        return json.load(data)


@lru_cache(maxsize=1, typed=False)
def default_dict() -> Converter:
    import pandas as pd

    return {
        "string": lambda col, df: df[col].astype(pd.StringDtype()) if hasattr(pd, "StringDType") else df[col],
        "guid": lambda col, df: df[col],
        "uuid": lambda col, df: df[col],
        "uniqueid": lambda col, df: df[col],
        "dynamic": lambda col, df: df[col],
        "bool": lambda col, df: df[col].astype(bool),
        "boolean": lambda col, df: df[col].astype(bool),
        "int": lambda col, df: df[col].astype(pd.Int32Dtype()),
        "int32": lambda col, df: df[col].astype(pd.Int32Dtype()),
        "int64": lambda col, df: df[col].astype(pd.Int64Dtype()),
        "long": lambda col, df: df[col].astype(pd.Int64Dtype()),
        "real": lambda col, df: parse_float(df, col),
        "double": lambda col, df: parse_float(df, col),
        "decimal": lambda col, df: parse_float(df, col),
        "datetime": lambda col, df: parse_datetime(df, col),
        "date": lambda col, df: parse_datetime(df, col),
        "timespan": lambda col, df: df[col].apply(parse_timedelta),
        "time": lambda col, df: df[col].apply(parse_timedelta),
    }


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License


def dataframe_from_result_table(
    table: "Union[KustoResultTable, KustoStreamingResultTable]",
    nullable_bools: bool = False,
    converters_by_type: Optional[Converter] = None,
    converters_by_column_name: Optional[Converter] = None,
) -> "pd.DataFrame":
    f"""Converts Kusto tables into pandas DataFrame.
    :param azure.kusto.data._models.KustoResultTable table: Table received from the response.
    :param nullable_bools: When True, converts bools that are 'null' from kusto or 'None' from python to pandas.NA. This will be the default in the future.
    :param converters_by_type: If given, converts specified types to corresponding types, else uses {default_dict()}. The dictionary maps from kusto
    datatype (https://learn.microsoft.com/azure/data-explorer/kusto/query/scalar-data-types/) to a lambda that receives a column name and a dataframe and
    returns the converted column or to a string type name.
    :param converters_by_column_name: If given, converts specified columns to corresponding types, else uses converters_by_type. The dictionary maps from column
     name to a lambda that receives a column name and a dataframe and returns the converted column.
    :return: pandas DataFrame.
    """
    import pandas as pd

    if not table:
        raise ValueError()

    from azure.kusto.data._models import KustoResultTable, KustoStreamingResultTable

    if not isinstance(table, KustoResultTable) and not isinstance(table, KustoStreamingResultTable):
        raise TypeError("Expected KustoResultTable or KustoStreamingResultTable got {}".format(type(table).__name__))

    columns = [col.column_name for col in table.columns]
    frame = pd.DataFrame(table.raw_rows, columns=columns)
    default = default_dict()

    for col in table.columns:
        column_name = col.column_name
        column_type = col.column_type
        if converters_by_column_name and column_name in converters_by_column_name:
            converter = converters_by_column_name.get(column_name)
        elif converters_by_type and column_type in converters_by_type:
            converter = converters_by_type.get(column_type)
        elif nullable_bools and column_type == "bool":
            converter = lambda col, df: df[col].astype(pd.BooleanDtype())
        else:
            converter = default.get(column_type)
        if converter is None:
            raise Exception("Unexpected type " + column_type)
        if isinstance(converter, str):
            frame[column_name] = frame[column_name].astype(converter)
        else:
            frame[column_name] = converter(column_name, frame)

    return frame


def get_string_tail_lower_case(val, length):
    if length <= 0:
        return ""

    if length >= len(val):
        return val.lower()

    return val[len(val) - length :].lower()


# TODO When moving to pandas 2 only - change to the appropriate type
def parse_float(frame, col):
    import numpy as np
    import pandas as pd

    frame[col] = frame[col].infer_objects(copy=False).replace({"NaN": np.nan, "Infinity": np.inf, "-Infinity": -np.inf})
    frame[col] = pd.to_numeric(frame[col], errors="coerce").astype(pd.Float64Dtype())  # pyright: ignore[reportCallIssue,reportArgumentType]

    return frame[col]


def parse_datetime(frame, col, force_version: Optional[str] = None) -> "pd.Series":
    # Pandas before version 2 doesn't support the "format" arg
    import pandas as pd

    args = {}
    if (force_version or pd.__version__).startswith("2."):
        args = {"format": "ISO8601", "utc": True}
    else:
        # if frame contains ".", replace "Z" with ".000Z"
        # Using bitwise NOT (~) on the boolean Series is the idiomatic pandas way to negate the mask
        contains_dot = frame[col].str.contains("\\.")
        frame.loc[~contains_dot, col] = frame.loc[~contains_dot, col].str.replace("Z", ".000Z")
    frame[col] = pd.to_datetime(frame[col], errors="coerce", **args)
    return frame[col]


def parse_timedelta(raw_value: Union[int, float, str]) -> "pd.Timedelta":
    """
    Transform a raw python value to a pandas timedelta.
    """
    import pandas as pd

    if isinstance(raw_value, (int, float)):
        # https://docs.microsoft.com/en-us/dotnet/api/system.datetime.ticks
        # Kusto saves up to ticks, 1 tick == 100 nanoseconds
        return pd.to_timedelta(raw_value * 100, unit="ns")
    if isinstance(raw_value, str):
        # The timespan format Kusto returns is 'd.hh:mm:ss.ssssss' or 'hh:mm:ss.ssssss' or 'hh:mm:ss'
        # Pandas expects 'd days hh:mm:ss.ssssss' or 'hh:mm:ss.ssssss' or 'hh:mm:ss'
        parts = raw_value.split(":")
        if "." not in parts[0]:
            return pd.to_timedelta(raw_value)
        else:
            formatted_value = raw_value.replace(".", " days ", 1)
            return pd.to_timedelta(formatted_value)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import re
from datetime import timedelta

from dateutil import parser

# Regex for TimeSpan
_TIMESPAN_PATTERN = re.compile(r"(-?)((?P<d>[0-9]*).)?(?P<h>[0-9]{2}):(?P<m>[0-9]{2}):(?P<s>[0-9]{2}(\.[0-9]+)?$)")


def to_datetime(value):
    """Converts a string to a datetime."""
    if isinstance(value, int):
        return parser.parse(value)
    return parser.isoparse(value)


def to_timedelta(value):
    """Converts a string to a timedelta."""
    if isinstance(value, (int, float)):
        return timedelta(microseconds=(float(value) / 10))
    match = _TIMESPAN_PATTERN.match(value)
    if match:
        if match.group(1) == "-":
            factor = -1
        else:
            factor = 1
        return factor * timedelta(days=int(match.group("d") or 0), hours=int(match.group("h")), minutes=int(match.group("m")), seconds=float(match.group("s")))
    else:
        raise ValueError("Timespan value '{}' cannot be decoded".format(value))



import io
from datetime import timedelta
from typing import Optional, Union

from azure.core.tracing import SpanKind
from azure.core.tracing.decorator_async import distributed_trace_async

from .response import KustoStreamingResponseDataSet
from .._decorators import aio_documented_by, documented_by
from .._telemetry import MonitoredActivity, Span
from ..aio.streaming_response import JsonTokenReader, StreamingDataSetEnumerator
from ..client import KustoClient as KustoClientSync
from ..client_base import ExecuteRequestParams, _KustoClientBase
from ..client_request_properties import ClientRequestProperties
from ..data_format import DataFormat
from ..exceptions import KustoAioSyntaxError, KustoClosedError, KustoNetworkError
from ..kcsb import KustoConnectionStringBuilder
from ..response import KustoResponseDataSet

try:
    from aiohttp import ClientResponse, ClientSession
except ImportError:
    raise KustoAioSyntaxError()


@documented_by(KustoClientSync)
class KustoClient(_KustoClientBase):
    @documented_by(KustoClientSync.__init__)
    def __init__(self, kcsb: Union[KustoConnectionStringBuilder, str]):
        super().__init__(kcsb, True)

        self._session = ClientSession()

    async def __aenter__(self) -> "KustoClient":
        return self

    async def close(self):
        if not self._is_closed:
            await self._session.close()
            if self._aad_helper:
                await self._aad_helper.close_async()
        super().close()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    @aio_documented_by(KustoClientSync.execute)
    async def execute(self, database: Optional[str], query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        query = query.strip()
        if query.startswith("."):
            return await self.execute_mgmt(database, query, properties)
        return await self.execute_query(database, query, properties)

    @distributed_trace_async(name_of_span="AioKustoClient.query_cmd", kind=SpanKind.CLIENT)
    @aio_documented_by(KustoClientSync.execute_query)
    async def execute_query(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        database = self._get_database_or_default(database)
        Span.set_query_attributes(self._kusto_cluster, database, properties)
        request = ExecuteRequestParams._from_query(
            query,
            database,
            properties,
            self._request_headers,
            self._query_default_timeout,
            self._mgmt_default_timeout,
            self._client_server_delta,
            self.client_details,
        )
        return await self._execute(self._query_endpoint, request, properties)

    @distributed_trace_async(name_of_span="AioKustoClient.control_cmd", kind=SpanKind.CLIENT)
    @aio_documented_by(KustoClientSync.execute_mgmt)
    async def execute_mgmt(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        database = self._get_database_or_default(database)
        Span.set_query_attributes(self._kusto_cluster, database, properties)
        request = ExecuteRequestParams._from_query(
            query,
            database,
            properties,
            self._request_headers,
            self._mgmt_default_timeout,
            self._mgmt_default_timeout,
            self._client_server_delta,
            self.client_details,
        )
        return await self._execute(self._mgmt_endpoint, request, properties)

    @distributed_trace_async(name_of_span="AioKustoClient.streaming_ingest", kind=SpanKind.CLIENT)
    @aio_documented_by(KustoClientSync.execute_streaming_ingest)
    async def execute_streaming_ingest(
        self,
        database: Optional[str],
        table: str,
        stream: Optional[io.IOBase],
        blob_url: Optional[str],
        stream_format: Union[DataFormat, str],
        properties: ClientRequestProperties = None,
        mapping_name: str = None,
    ):
        database = self._get_database_or_default(database)

        stream_format = stream_format.kusto_value if isinstance(stream_format, DataFormat) else DataFormat[stream_format.upper()].kusto_value
        endpoint = self._streaming_ingest_endpoint + database + "/" + table + "?streamFormat=" + stream_format
        if mapping_name is not None:
            endpoint = endpoint + "&mappingName=" + mapping_name

        if blob_url:
            endpoint += "&sourceKind=uri"
            request = ExecuteRequestParams._from_blob_url(
                blob_url,
                properties,
                self._request_headers,
                self._streaming_ingest_default_timeout,
                self._mgmt_default_timeout,
                self._client_server_delta,
                self.client_details,
            )
        elif stream:
            request = ExecuteRequestParams._from_stream(
                stream,
                properties,
                self._request_headers,
                self._streaming_ingest_default_timeout,
                self._mgmt_default_timeout,
                self._client_server_delta,
                self.client_details,
            )
        else:
            raise Exception("execute_streaming_ingest is expecting either a stream or blob url")

        Span.set_streaming_ingest_attributes(self._kusto_cluster, database, table, properties)
        await self._execute(endpoint, request, properties)

    @aio_documented_by(KustoClientSync._execute_streaming_query_parsed)
    async def _execute_streaming_query_parsed(
        self,
        database: Optional[str],
        query: str,
        timeout: timedelta = _KustoClientBase._query_default_timeout,
        properties: Optional[ClientRequestProperties] = None,
    ) -> StreamingDataSetEnumerator:
        request = ExecuteRequestParams._from_query(
            query, database, properties, self._request_headers, timeout, self._mgmt_default_timeout, self._client_server_delta, self.client_details
        )
        response = await self._execute(self._query_endpoint, request, properties, stream_response=True)
        return StreamingDataSetEnumerator(JsonTokenReader(response.content))

    @distributed_trace_async(name_of_span="AioKustoClient.streaming_query", kind=SpanKind.CLIENT)
    @aio_documented_by(KustoClientSync.execute_streaming_query)
    async def execute_streaming_query(
        self,
        database: Optional[str],
        query: str,
        timeout: timedelta = _KustoClientBase._query_default_timeout,
        properties: Optional[ClientRequestProperties] = None,
    ) -> KustoStreamingResponseDataSet:
        database = self._get_database_or_default(database)
        Span.set_query_attributes(self._kusto_cluster, database, properties)

        response = await self._execute_streaming_query_parsed(database, query, timeout, properties)
        return KustoStreamingResponseDataSet(response)

    @aio_documented_by(KustoClientSync._execute)
    async def _execute(
        self,
        endpoint: str,
        request: ExecuteRequestParams,
        properties: Optional[ClientRequestProperties] = None,
        stream_response: bool = False,
    ) -> Union[KustoResponseDataSet, ClientResponse]:
        """Executes given query against this client"""
        if self._is_closed:
            raise KustoClosedError()
        self.validate_endpoint()

        request_headers = request.request_headers
        timeout = request.timeout
        if self._aad_helper:
            request_headers["Authorization"] = await self._aad_helper.acquire_authorization_header_async()

        invoker = lambda: self._session.post(
            endpoint,
            headers=request_headers,
            json=request.json_payload,
            data=request.payload,
            timeout=timeout.seconds,
            proxy=self._proxy_url,
            allow_redirects=False,
        )

        try:
            response = await MonitoredActivity.invoke_async(
                invoker, name_of_span="AioKustoClient.http_post", tracing_attributes=Span.create_http_attributes("POST", endpoint, request_headers)
            )
        except Exception as e:
            raise KustoNetworkError(endpoint, None if properties is None else properties.client_request_id) from e

        if stream_response:
            try:
                response.raise_for_status()
                if 300 <= response.status < 400:
                    raise Exception("Unexpected redirection, got status code: " + str(response.status))
                return response
            except Exception as e:
                try:
                    response_text = await response.text()
                except Exception:
                    response_text = None
                try:
                    response_json = await response.json()
                except Exception:
                    response_json = None
                raise self._handle_http_error(e, endpoint, request.payload, response, response.status, response_json, response_text)

        async with response:
            response_json = None
            try:
                if 300 <= response.status < 400:
                    raise Exception("Unexpected redirection, got status code: " + str(response.status))
                response_json = await response.json()
                response.raise_for_status()
            except Exception as e:
                try:
                    response_text = await response.text()
                except Exception:
                    response_text = None
                raise self._handle_http_error(e, endpoint, request.payload, response, response.status, response_json, response_text)
            return MonitoredActivity.invoke(lambda: self._kusto_parse_by_endpoint(endpoint, response_json), name_of_span="AioKustoClient.processing_response")



import dataclasses
from threading import Lock
from typing import Optional, Dict
from urllib.parse import urlparse

import requests

from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing import SpanKind

from .env_utils import get_env
from ._telemetry import Span, MonitoredActivity
from .exceptions import KustoServiceError, KustoNetworkError

METADATA_ENDPOINT = "v1/rest/auth/metadata"

DEFAULT_AUTH_ENV_VAR_NAME = "AadAuthorityUri"
DEFAULT_KUSTO_CLIENT_APP_ID = "db662dc1-0cfe-4e1c-a843-19a68e65be58"
DEFAULT_PUBLIC_LOGIN_URL = "https://login.microsoftonline.com"
DEFAULT_REDIRECT_URI = "http://localhost"
DEFAULT_KUSTO_SERVICE_RESOURCE_ID = "https://kusto.kusto.windows.net"
DEFAULT_DEV_KUSTO_SERVICE_RESOURCE_ID = "https://kusto.dev.kusto.windows.net"
DEFAULT_FIRST_PARTY_AUTHORITY_URL = "https://login.microsoftonline.com/f8cdef31-a31e-4b4a-93e4-5f571e91255a"


@dataclasses.dataclass
class CloudInfo:
    """This class holds the data for a specific cloud instance."""

    login_endpoint: str
    login_mfa_required: bool
    kusto_client_app_id: str
    kusto_client_redirect_uri: str
    kusto_service_resource_id: str
    first_party_authority_url: str

    def authority_uri(self, authority_id: Optional[str]):
        return self.login_endpoint + "/" + (authority_id or "organizations")


class CloudSettings:
    """This class holds data for all cloud instances, and returns the specific data instance by parsing the dns suffix from a URL"""

    _cloud_info = None
    _cloud_cache = {}
    _cloud_cache_lock = Lock()

    DEFAULT_CLOUD = CloudInfo(
        login_endpoint=get_env(DEFAULT_AUTH_ENV_VAR_NAME, default=DEFAULT_PUBLIC_LOGIN_URL),
        login_mfa_required=False,
        kusto_client_app_id=DEFAULT_KUSTO_CLIENT_APP_ID,
        kusto_client_redirect_uri=DEFAULT_REDIRECT_URI,
        kusto_service_resource_id=DEFAULT_KUSTO_SERVICE_RESOURCE_ID,
        first_party_authority_url=DEFAULT_FIRST_PARTY_AUTHORITY_URL,
    )

    @classmethod
    @distributed_trace(name_of_span="CloudSettings.get_cloud_info", kind=SpanKind.CLIENT)
    def get_cloud_info_for_cluster(cls, kusto_uri: str, proxies: Optional[Dict[str, str]] = None, session: requests.Session = None) -> CloudInfo:
        normalized_authority = cls._normalize_uri(kusto_uri)

        # tracing attributes for cloud info
        Span.set_cloud_info_attributes(kusto_uri)

        if normalized_authority in cls._cloud_cache:  # Double-checked locking to avoid unnecessary lock access
            return cls._cloud_cache[normalized_authority]

        with cls._cloud_cache_lock:
            if normalized_authority in cls._cloud_cache:
                return cls._cloud_cache[normalized_authority]

            url_parts = urlparse(kusto_uri)
            url = f"{url_parts.scheme}://{url_parts.netloc}/{METADATA_ENDPOINT}"

            try:
                # trace http get call for result
                result = MonitoredActivity.invoke(
                    lambda: (session or requests).get(url, proxies=proxies, allow_redirects=False),
                    name_of_span="CloudSettings.http_get",
                    tracing_attributes=Span.create_http_attributes(url=url, method="GET"),
                )
            except Exception as e:
                raise KustoNetworkError(url) from e

            if result.status_code == 200:
                content = result.json()
                if content is None or content == {}:
                    raise KustoServiceError("Kusto returned an invalid cloud metadata response", result)
                root = content["AzureAD"]
                if root is not None:
                    cls._cloud_cache[normalized_authority] = CloudInfo(
                        login_endpoint=root["LoginEndpoint"],
                        login_mfa_required=root["LoginMfaRequired"],
                        kusto_client_app_id=root["KustoClientAppId"],
                        kusto_client_redirect_uri=root["KustoClientRedirectUri"],
                        kusto_service_resource_id=root["KustoServiceResourceId"],
                        first_party_authority_url=root["FirstPartyAuthorityUrl"],
                    )
                else:
                    cls._cloud_cache[normalized_authority] = cls.DEFAULT_CLOUD
            elif result.status_code == 404:
                # For now as long not all proxies implement the metadata endpoint, if no endpoint exists return public cloud data
                cls._cloud_cache[normalized_authority] = cls.DEFAULT_CLOUD
            else:
                raise KustoServiceError("Kusto returned an invalid cloud metadata response", result)
            return cls._cloud_cache[normalized_authority]

    @classmethod
    def add_to_cache(cls, url: str, cloud_info: CloudInfo):
        with cls._cloud_cache_lock:
            cls._cloud_cache[cls._normalize_uri(url)] = cloud_info

    @classmethod
    def _normalize_uri(cls, kusto_uri):
        """Extracts and returns the authority part of the URI (schema, host, port)"""
        url_parts = urlparse(kusto_uri)
        # Return only the scheme and netloc (which contains host and port if present)
        return f"{url_parts.scheme}://{url_parts.netloc}"



import dataclasses
from enum import unique, Enum
from typing import Union, Callable, Coroutine, Optional, Tuple, List, Any, ClassVar
from urllib.parse import urlparse

from ._string_utils import assert_string_is_not_empty
from ._token_providers import DeviceCallbackType
from .client_details import ClientDetails
from .helpers import load_bundled_json


UNSUPPORTED_KEYWORD = "UNSUPPORTED"


@unique
class SupportedKeywords(Enum):
    DATA_SOURCE = "Data Source"
    INITIAL_CATALOG = "Initial Catalog"
    FEDERATED_SECURITY = "AAD Federated Security"
    APPLICATION_CLIENT_ID = "Application Client Id"
    APPLICATION_KEY = "Application Key"
    USER_ID = "User ID"
    PASSWORD = "Password"
    AUTHORITY_ID = "Authority Id"
    APPLICATION_TOKEN = "Application Token"
    USER_TOKEN = "User Token"
    APPLICATION_CERTIFICATE_BLOB = "Application Certificate Blob"
    APPLICATION_CERTIFICATE_X5C = "Application Certificate SendX5c"
    APPLICATION_CERTIFICATE_THUMBPRINT = "Application Certificate Thumbprint"
    TRACE_APP_NAME = "Application Name for Tracing"
    TRACE_USER_NAME = "User Name for Tracing"


@unique
class UnsupportedKeywords(Enum):
    DSTS_FEDERATED_SECURITY = "dSTS Federated Security"
    STREAMING = "Streaming"
    UNCOMPRESSED = "Uncompressed"
    ENFORCE_MFA = "EnforceMfa"
    ACCEPT = "Accept"
    QUERY_CONSISTENCY = "Query Consistency"
    DATA_SOURCE_URI = "Data Source Uri"
    AZURE_REGION = "Azure Region"
    NAMESPACE = "Namespace"
    APPLICATION_CERTIFICATE_ISSUER_DISTINGUISHED_NAME = "Application Certificate Issuer Distinguished Name"
    APPLICATION_CERTIFICATE_SUBJECT_DISTINGUISHED_NAME = "Application Certificate Subject Distinguished Name"


@dataclasses.dataclass(frozen=True)
class Keyword:
    _supported_keywords: ClassVar[List[str]] = [k.value for k in SupportedKeywords]
    _unsupported_keywords: ClassVar[List[str]] = [k.value for k in UnsupportedKeywords]
    _lookup: ClassVar[dict]

    name: SupportedKeywords
    type: str
    secret: bool

    def is_str_type(self) -> bool:
        return self.type == "string"

    def is_bool_type(self) -> bool:
        return self.type == "bool"

    @staticmethod
    def normalize_string(key: str) -> str:
        return key.lower().replace(" ", "")

    @classmethod
    def init_lookup(cls):
        kcsb_json: dict = load_bundled_json("kcsb.json")
        lookup = {}
        for v in kcsb_json["keywords"]:
            name = v["name"]
            if name in cls._supported_keywords:
                keyword = Keyword(SupportedKeywords(name), v["type"], v["secret"])
            elif name in cls._unsupported_keywords:
                keyword = UNSUPPORTED_KEYWORD
            else:
                raise KeyError(f"Unknown keyword: `{name}`")

            lookup[Keyword.normalize_string(name)] = keyword

            for alias in v["aliases"]:
                lookup[Keyword.normalize_string(alias)] = keyword

        cls._lookup = lookup

    @classmethod
    def parse(cls, key: Union[str, SupportedKeywords]) -> "Keyword":
        if isinstance(key, SupportedKeywords):
            key = key.value

        normalized = Keyword.normalize_string(key)

        if normalized not in cls._lookup:
            raise KeyError(f"Unknown keyword: `{key}`")

        if cls._lookup[normalized] == UNSUPPORTED_KEYWORD:
            raise KeyError(f"Keyword `{key}` is not supported by this SDK")

        return cls._lookup[normalized]

    @classmethod
    def lookup(cls, key: Union[str, SupportedKeywords]) -> "Keyword":
        if isinstance(key, SupportedKeywords):
            key = key.value

        return cls._lookup[Keyword.normalize_string(key)]


Keyword.init_lookup()


class KustoConnectionStringBuilder:
    """
    Parses Kusto connection strings.
    For usages, check out the sample at:
        https://github.com/Azure/azure-kusto-python/blob/master/azure-kusto-data/tests/sample.py
    """

    DEFAULT_DATABASE_NAME = "NetDefaultDB"

    interactive_login: bool = False
    az_cli_login: bool = False
    device_login: bool = False
    token_credential_login: bool = False

    device_callback: DeviceCallbackType = None
    msi_authentication: bool = False
    msi_parameters: Optional[dict] = None

    token_provider: Optional[Callable[[], str]] = None
    async_token_provider: Optional[Callable[[], Coroutine[None, None, str]]] = None

    application_for_tracing: Optional[str] = None
    user_name_for_tracing: Optional[str] = None

    azure_credential: Optional[Any] = None
    azure_credential_from_login_endpoint: Optional[Any] = None

    application_public_certificate: Optional[str] = None

    def __init__(self, connection_string: str):
        """
        Creates new KustoConnectionStringBuilder.
        :param str connection_string: Kusto connection string should be of the format:
        https://<clusterName>.kusto.windows.net;AAD User ID="user@microsoft.com";Password=P@ssWord
        For more information please look at:
        https://kusto.azurewebsites.net/docs/concepts/kusto_connection_strings.html
        """
        assert_string_is_not_empty(connection_string)
        self._internal_dict = {}

        if connection_string is not None and "=" not in connection_string.partition(";")[0]:
            connection_string = "Data Source=" + connection_string

        self[SupportedKeywords.AUTHORITY_ID] = "organizations"

        for kvp_string in connection_string.split(";"):
            key, _, value = kvp_string.partition("=")
            keyword = Keyword.parse(key)

            value_stripped = value.strip()
            if keyword.is_str_type():
                if keyword.name == SupportedKeywords.DATA_SOURCE:
                    self[keyword.name] = value_stripped.rstrip("/")
                    self._parse_data_source(self.data_source)
                elif keyword.name == SupportedKeywords.TRACE_USER_NAME:
                    self.user_name_for_tracing = value_stripped
                elif keyword.name == SupportedKeywords.TRACE_APP_NAME:
                    self.application_for_tracing = value_stripped
                else:
                    self[keyword.name] = value_stripped
            elif keyword.is_bool_type():
                if value_stripped in ["True", "true"]:
                    self[keyword.name] = True
                elif value_stripped in ["False", "false"]:
                    self[keyword.name] = False
                else:
                    raise KeyError("Expected aad federated security to be bool. Recieved %s" % value)

        if self.initial_catalog is None:
            self.initial_catalog = self.DEFAULT_DATABASE_NAME

    def __setitem__(self, key: "Union[SupportedKeywords, str]", value: Union[str, bool, dict]):
        keyword = Keyword.parse(key)

        if value is None:
            raise TypeError("Value cannot be None.")

        if keyword.is_str_type():
            self._internal_dict[keyword.name] = value.strip()
        elif keyword.is_bool_type():
            if not isinstance(value, bool):
                raise TypeError("Expected %s to be bool" % key)
            self._internal_dict[keyword.name] = value
        else:
            raise KeyError("KustoConnectionStringBuilder supports only bools and strings.")

    @classmethod
    def with_aad_user_password_authentication(
        cls, connection_string: str, user_id: str, password: str, authority_id: str = "organizations"
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD user name and
        password.
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param str user_id: AAD user ID.
        :param str password: Corresponding password of the AAD user.
        :param str authority_id: optional param. defaults to "organizations"
        """
        assert_string_is_not_empty(user_id)
        assert_string_is_not_empty(password)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.USER_ID] = user_id
        kcsb[SupportedKeywords.AUTHORITY_ID] = authority_id
        kcsb[SupportedKeywords.PASSWORD] = password

        return kcsb

    @classmethod
    def with_aad_user_token_authentication(cls, connection_string: str, user_token: str) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application and
        a certificate credentials.
        :param str connection_string: Kusto connection string should be of the format:
        https://<clusterName>.kusto.windows.net
        :param str user_token: AAD user token.
        """
        assert_string_is_not_empty(user_token)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.USER_TOKEN] = user_token

        return kcsb

    @classmethod
    def with_aad_application_key_authentication(
        cls, connection_string: str, aad_app_id: str, app_key: str, authority_id: str
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application and key.
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param str aad_app_id: AAD application ID.
        :param str app_key: Corresponding key of the AAD application.
        :param str authority_id: Authority id (aka Tenant id) must be provided
        """
        assert_string_is_not_empty(aad_app_id)
        assert_string_is_not_empty(app_key)
        assert_string_is_not_empty(authority_id)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.APPLICATION_CLIENT_ID] = aad_app_id
        kcsb[SupportedKeywords.APPLICATION_KEY] = app_key
        kcsb[SupportedKeywords.AUTHORITY_ID] = authority_id

        return kcsb

    @classmethod
    def with_aad_application_certificate_authentication(
        cls, connection_string: str, aad_app_id: str, certificate: str, thumbprint: str, authority_id: str
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application using
        a certificate.
        :param str connection_string: Kusto connection string should be of the format:
        https://<clusterName>.kusto.windows.net
        :param str aad_app_id: AAD application ID.
        :param str certificate: A PEM encoded certificate private key.
        :param str thumbprint: hex encoded thumbprint of the certificate.
        :param str authority_id: Authority id (aka Tenant id) must be provided
        """
        assert_string_is_not_empty(aad_app_id)
        assert_string_is_not_empty(certificate)
        assert_string_is_not_empty(thumbprint)
        assert_string_is_not_empty(authority_id)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.APPLICATION_CLIENT_ID] = aad_app_id
        kcsb[SupportedKeywords.APPLICATION_CERTIFICATE_BLOB] = certificate
        kcsb[SupportedKeywords.APPLICATION_CERTIFICATE_THUMBPRINT] = thumbprint
        kcsb[SupportedKeywords.AUTHORITY_ID] = authority_id

        return kcsb

    @classmethod
    def with_aad_application_certificate_sni_authentication(
        cls, connection_string: str, aad_app_id: str, private_certificate: str, public_certificate: str, thumbprint: str, authority_id: str
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application using
        a certificate Subject Name and Issuer.
        :param str connection_string: Kusto connection string should be of the format:
        https://<clusterName>.kusto.windows.net
        :param str aad_app_id: AAD application ID.
        :param str private_certificate: A PEM encoded certificate private key.
        :param str public_certificate: A public certificate matching the provided PEM certificate private key.
        :param str thumbprint: hex encoded thumbprint of the certificate.
        :param str authority_id: Authority id (aka Tenant id) must be provided
        """
        assert_string_is_not_empty(aad_app_id)
        assert_string_is_not_empty(private_certificate)
        assert_string_is_not_empty(public_certificate)
        assert_string_is_not_empty(thumbprint)
        assert_string_is_not_empty(authority_id)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.APPLICATION_CLIENT_ID] = aad_app_id
        kcsb[SupportedKeywords.APPLICATION_CERTIFICATE_BLOB] = private_certificate
        kcsb.application_public_certificate = public_certificate
        kcsb[SupportedKeywords.APPLICATION_CERTIFICATE_THUMBPRINT] = thumbprint
        kcsb[SupportedKeywords.AUTHORITY_ID] = authority_id

        return kcsb

    @classmethod
    def with_aad_application_token_authentication(cls, connection_string: str, application_token: str) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application and
        an application token.
        :param str connection_string: Kusto connection string should be of the format:
        https://<clusterName>.kusto.windows.net
        :param str application_token: AAD application token.
        """
        assert_string_is_not_empty(application_token)
        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.APPLICATION_TOKEN] = application_token

        return kcsb

    @classmethod
    def with_aad_device_authentication(
        cls, connection_string: str, authority_id: str = "organizations", callback: DeviceCallbackType = None
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application and
        password.
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param str authority_id: optional param. defaults to "organizations"
        :param DeviceCallbackType callback: options callback function to be called when authentication is required, accepts three parameters:
                - ``verification_uri`` (str) the URL the user must visit
                - ``user_code`` (str) the code the user must enter there
                - ``expires_on`` (datetime.datetime) the UTC time at which the code will expire
        """
        kcsb = cls(connection_string)
        kcsb.device_login = True
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb[SupportedKeywords.AUTHORITY_ID] = authority_id
        kcsb.device_callback = callback

        return kcsb

    @classmethod
    def with_az_cli_authentication(cls, connection_string: str) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will use existing authenticated az cli profile
        password.
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        """
        kcsb = cls(connection_string)
        kcsb.az_cli_login = True
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True

        return kcsb

    @classmethod
    def with_aad_managed_service_identity_authentication(
        cls, connection_string: str, client_id: str = None, object_id: str = None, msi_res_id: str = None, timeout: int = None
    ) -> "KustoConnectionStringBuilder":
        """
        Creates a KustoConnection string builder that will authenticate with AAD application, using
        an application token obtained from a Microsoft Service Identity endpoint. An optional user
        assigned application ID can be added to the token.

        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param client_id: an optional user assigned identity provided as an Azure ID of a client
        :param object_id: an optional user assigned identity provided as an Azure ID of an object
        :param msi_res_id: an optional user assigned identity provided as an Azure ID of an MSI resource
        :param timeout: an optional timeout (seconds) to wait for an MSI Authentication to occur
        """

        kcsb = cls(connection_string)
        params = {}
        exclusive_pcount = 0

        if timeout is not None:
            params["connection_timeout"] = timeout

        if client_id is not None:
            params["client_id"] = client_id
            exclusive_pcount += 1

        if object_id is not None:
            # Until we upgrade azure-identity to version 1.4.1, only client_id is excepted as a hint for user managed service identity
            raise ValueError("User Managed Service Identity with object_id is temporarily not supported by azure identity 1.3.1. Please use client_id instead.")
            # noinspection PyUnreachableCode
            params["object_id"] = object_id
            exclusive_pcount += 1

        if msi_res_id is not None:
            # Until we upgrade azure-identity to version 1.4.1, only client_id is excepted as a hint for user managed service identity
            raise ValueError(
                "User Managed Service Identity with msi_res_id is temporarily not supported by azure identity 1.3.1. Please use client_id instead."
            )
            # noinspection PyUnreachableCode
            params["msi_res_id"] = msi_res_id
            exclusive_pcount += 1

        if exclusive_pcount > 1:
            raise ValueError("the following parameters are mutually exclusive and can not be provided at the same time: client_uid, object_id, msi_res_id")

        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb.msi_authentication = True
        kcsb.msi_parameters = params

        return kcsb

    @classmethod
    def with_token_provider(cls, connection_string: str, token_provider: Callable[[], str]) -> "KustoConnectionStringBuilder":
        """
        Create a KustoConnectionStringBuilder that uses a callback function to obtain a connection token
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param token_provider: a parameterless function that returns a valid bearer token for the relevant kusto resource as a string
        """

        assert callable(token_provider)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb.token_provider = token_provider

        return kcsb

    @classmethod
    def with_async_token_provider(
        cls,
        connection_string: str,
        async_token_provider: Callable[[], Coroutine[None, None, str]],
    ) -> "KustoConnectionStringBuilder":
        """
        Create a KustoConnectionStringBuilder that uses an async callback function to obtain a connection token
        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param async_token_provider: a parameterless function that after awaiting returns a valid bearer token for the relevant kusto resource as a string
        """

        assert callable(async_token_provider)

        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb.async_token_provider = async_token_provider

        return kcsb

    @classmethod
    def with_interactive_login(
        cls, connection_string: str, user_id_hint: Optional[str] = None, tenant_hint: Optional[str] = None
    ) -> "KustoConnectionStringBuilder":
        kcsb = cls(connection_string)
        kcsb.interactive_login = True
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        if user_id_hint is not None:
            kcsb[SupportedKeywords.USER_ID] = user_id_hint

        if tenant_hint is not None:
            kcsb[SupportedKeywords.AUTHORITY_ID] = tenant_hint

        return kcsb

    @classmethod
    def with_azure_token_credential(
        cls,
        connection_string: str,
        credential: Optional[Any] = None,
        credential_from_login_endpoint: Optional[Callable[[str], Any]] = None,
    ) -> "KustoConnectionStringBuilder":
        """
        Create a KustoConnectionStringBuilder that uses an azure token credential to obtain a connection token.
        :param connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
        :param credential: an optional token credential to use for authentication
        :param credential_from_login_endpoint: an optional function that returns a token credential for the relevant kusto resource
        """
        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = True
        kcsb.token_credential_login = True
        kcsb.azure_credential = credential
        kcsb.azure_credential_from_login_endpoint = credential_from_login_endpoint

        return kcsb

    @classmethod
    def with_no_authentication(cls, connection_string: str) -> "KustoConnectionStringBuilder":
        """
        Create a KustoConnectionStringBuilder that uses no authentication.
        :param connection_string: Kusto's connection string should be of the format: http://<clusterName>.kusto.windows.net
        """
        if not connection_string.startswith("http://"):
            raise ValueError("Connection string must start with http://")
        kcsb = cls(connection_string)
        kcsb[SupportedKeywords.FEDERATED_SECURITY] = False

        return kcsb

    @property
    def data_source(self) -> Optional[str]:
        """The URI specifying the Kusto service endpoint.
        For example, https://kuskus.kusto.windows.net or net.tcp://localhost
        """
        return self._internal_dict.get(SupportedKeywords.DATA_SOURCE)

    @property
    def initial_catalog(self) -> Optional[str]:
        """The default database to be used for requests.
        By default, it is set to 'NetDefaultDB'.
        """
        return self._internal_dict.get(SupportedKeywords.INITIAL_CATALOG)

    @initial_catalog.setter
    def initial_catalog(self, value: str) -> None:
        self._internal_dict[SupportedKeywords.INITIAL_CATALOG] = value

    @property
    def aad_user_id(self) -> Optional[str]:
        """The username to use for AAD Federated AuthN."""
        return self._internal_dict.get(SupportedKeywords.USER_ID)

    @property
    def application_client_id(self) -> Optional[str]:
        """The application client id to use for authentication when federated
        authentication is used.
        """
        return self._internal_dict.get(SupportedKeywords.APPLICATION_CLIENT_ID)

    @property
    def application_key(self) -> Optional[str]:
        """The application key to use for authentication when federated authentication is used"""
        return self._internal_dict.get(SupportedKeywords.APPLICATION_KEY)

    @property
    def application_certificate(self) -> Optional[str]:
        """A PEM encoded certificate private key."""
        return self._internal_dict.get(SupportedKeywords.APPLICATION_CERTIFICATE_BLOB)

    @application_certificate.setter
    def application_certificate(self, value: str):
        self[SupportedKeywords.APPLICATION_CERTIFICATE_BLOB] = value

    @property
    def application_certificate_thumbprint(self) -> Optional[str]:
        """hex encoded thumbprint of the certificate."""
        return self._internal_dict.get(SupportedKeywords.APPLICATION_CERTIFICATE_THUMBPRINT)

    @application_certificate_thumbprint.setter
    def application_certificate_thumbprint(self, value: str):
        self[SupportedKeywords.APPLICATION_CERTIFICATE_THUMBPRINT] = value

    @property
    def authority_id(self) -> Optional[str]:
        """The ID of the AAD tenant where the application is configured.
        (should be supplied only for non-Microsoft tenant)"""
        return self._internal_dict.get(SupportedKeywords.AUTHORITY_ID)

    @authority_id.setter
    def authority_id(self, value: str):
        self[SupportedKeywords.AUTHORITY_ID] = value

    @property
    def aad_federated_security(self) -> Optional[bool]:
        """A Boolean value that instructs the client to perform AAD federated authentication."""
        return self._internal_dict.get(SupportedKeywords.FEDERATED_SECURITY)

    @property
    def user_token(self) -> Optional[str]:
        """User token."""
        return self._internal_dict.get(SupportedKeywords.USER_TOKEN)

    @property
    def application_token(self) -> Optional[str]:
        """Application token."""
        return self._internal_dict.get(SupportedKeywords.APPLICATION_TOKEN)

    @property
    def client_details(self) -> ClientDetails:
        return ClientDetails(self.application_for_tracing, self.user_name_for_tracing)

    @property
    def login_hint(self) -> Optional[str]:
        return self._internal_dict.get(SupportedKeywords.USER_ID)

    @property
    def domain_hint(self) -> Optional[str]:
        return self._internal_dict.get(SupportedKeywords.AUTHORITY_ID)

    @property
    def password(self) -> Optional[str]:
        return self._internal_dict.get(SupportedKeywords.PASSWORD)

    def _set_connector_details(
        self,
        name: str,
        version: str,
        app_name: Optional[str] = None,
        app_version: Optional[str] = None,
        send_user: bool = False,
        override_user: Optional[str] = None,
        additional_fields: Optional[List[Tuple[str, str]]] = None,
    ):
        """
        Sets the connector details for tracing purposes.
        :param name:  The name of the connector
        :param version:  The version of the connector
        :param send_user: Whether to send the user name
        :param override_user: Override the user name ( if send_user is True )
        :param app_name: The name of the containing application
        :param app_version: The version of the containing application
        :param additional_fields: Additional fields to add to the header
        """
        client_details = ClientDetails.set_connector_details(name, version, app_name, app_version, send_user, override_user, additional_fields)

        self.application_for_tracing = client_details.application_for_tracing
        self.user_name_for_tracing = client_details.user_name_for_tracing

    def __str__(self) -> str:
        dict_copy = self._internal_dict.copy()
        for key in dict_copy:
            if Keyword.lookup(key).secret:
                dict_copy[key] = "****"
        return self._build_connection_string(dict_copy)

    def __repr__(self) -> str:
        return self._build_connection_string(self._internal_dict)

    def _build_connection_string(self, kcsb_as_dict: dict) -> str:
        return ";".join(["{0}={1}".format(word.value, kcsb_as_dict[word]) for word in SupportedKeywords if word in kcsb_as_dict])

    def _parse_data_source(self, url: str):
        url = urlparse(url)
        if not url.netloc:
            return
        segments = url.path.lstrip("/").split("/")
        if len(segments) == 1 and segments[0] and not self.initial_catalog:
            self.initial_catalog = segments[0]
            self._internal_dict[SupportedKeywords.DATA_SOURCE] = url._replace(path="").geturl()
